package org.processmining.plugins.ltlchecker;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XTrace;
import org.omwg.ontology.Concept;
import org.omwg.ontology.Instance;
import org.wsmo.common.exception.InvalidModelException;
import org.wsmo.common.exception.SynchronisationException;

public class ConceptModel implements Comparable<ConceptModel> {
	private final Concept concept;
	private final OntologyModel ontology;
	private final Set<String> instances;

	public final static int WORKFLOW_MODEL_ELEMENT = 1;
	public final static int EVENTTYPE = 2;
	public final static int ORIGINATOR = 4;

	// has package visibility, so it can only be used by OntologyModel
	ConceptModel(OntologyModel ontology, Concept concept, Set<String> instances) {
		this.ontology = ontology;
		this.concept = concept;
		this.instances = instances == null ? new HashSet<String>(0) : instances;
	}

	public OntologyModel getOntology() {
		return ontology;
	}

	public String getName() {
		return concept.getIdentifier().toString();
	}

	public String getShortName() {
		return OntologyModel.getConceptPart(getName());
	}

	public String toString() {
		return getName();
	}

	@Override
	public int hashCode() {
		return getName().hashCode();
	}

	@Override
	public boolean equals(Object other) {
		if (!(other instanceof ConceptModel)) {
			return false;
		}
		return ((ConceptModel) other).getName().equals(getName());
	}

	Concept getConcept() {
		return concept;
	}

	public Set<String> getInstances() {
		return instances;
	}

	public int compareTo(ConceptModel other) {
		return getName().compareTo((other).getName());
	}

	public void addSuperConcept(ConceptModel superConcept) throws InvalidModelException {
		try {
			getConcept().addSuperConcept(superConcept.getConcept());
			ontology.structureChanged();
		} catch (SynchronisationException e) {
			throw new InvalidModelException(e.getMessage());
		} catch (InvalidModelException e) {
			throw new InvalidModelException(e.getMessage());
		}
	}

	/**
	 * Add ATE elements as instances to this concept and annotate the ATE with
	 * this concept.
	 * 
	 * @param pi
	 *            The process instance of the ATE to annotate
	 * @param index
	 *            The index in the PI of the ATE to annotate
	 * @param elementType
	 *            The element(s) to annotate: WORKFLOW_MODEL_ELEMENT, EVENTTYPE
	 *            or ORIGINATOR.
	 * @throws IndexOutOfBoundsException
	 * @throws IOException
	 * @throws org.processmining.plugins.ltlchecker.models.ontology.InvalidModelException
	 */
	public void addInstance(XEvent ate, int elementType) throws IOException, InvalidModelException {
		switch (elementType) {
			case WORKFLOW_MODEL_ELEMENT :
				//newRefs = addInstance(ate.getElementModelReferences(), ate.getElement());
				//ate.setElementModelReferences(newRefs);
				break;
			case ORIGINATOR :
				//newRefs = addInstance(ate.getOriginatorModelReferences(), ate.getOriginator());
				//	ate.setOriginatorModelReferences(newRefs);
				break;
			case EVENTTYPE :
				//	newRefs = addInstance(ate.getTypeModelReferences(), ate.getType());
				//	ate.setTypeModelReferences(newRefs);
				break;
			default :
				throw new IllegalArgumentException("Invalid elementType in call to ConceptModel.addInstance");
		}
	}

	public void addInstance(XTrace pi) throws InvalidModelException {
		//pi.setModelReferences(
		//	addInstance(pi.getModelReferences(), pi.getName())
		//		);
	}

	public void addInstance(Process process) throws InvalidModelException {
		//	process.setModelReferences(
		//	addInstance(process.getModelReferences(), process.getName())
		//			);
	}

	public void addInstance(InfoItem item) throws InvalidModelException {
		item.setModelReferences(addInstance(item.getModelReferences(), item.getName()));
	}

	public void addInstance(DataSection dataAttributes, String key) throws InvalidModelException {
		dataAttributes.setModelReferences(key,
				addInstance(dataAttributes.getModelReferences(key), dataAttributes.get(key)));
	}

	private List<String> addInstance(List<String> inputModelRefs, String instanceName) throws InvalidModelException {
		if (inputModelRefs.contains(instanceName)) { // TODO this is a linear search now!
			return inputModelRefs;
		}

		List<String> refs = new ArrayList<String>(inputModelRefs);

		refs.add(getName());

		addInstance(sanitizeName(instanceName));
		ontology.setDirectlyReferenced(this);
		return refs;
	}

	public void addInstance(String instanceName) {
		String uri = OntologyModel.getOntologyPart(ontology.getName()) + OntologyModel.ONTOLOGY_SEPARATOR
				+ instanceName;

		try {
			Instance instance = ontology.getFactory().createInstance(ontology.getFactory().createIRI(uri), concept);
			ontology.getWsmoOntology().addInstance(instance);
			instances.add(instanceName);
		} catch (SynchronisationException e) {
			e.printStackTrace();
		} catch (InvalidModelException e) {
			e.printStackTrace();
		}
	}

	private static boolean isValidChar(char ch) {
		return isValidFirstChar(ch) || (('0' <= ch) && (ch <= '9'));
	}

	private static boolean isValidFirstChar(char ch) {
		return (ch == '_') || (('a' <= ch) && (ch <= 'z')) || (('A' <= ch) && (ch <= 'Z'));
	}

	public static String sanitizeName(String name) {
		StringBuffer result = new StringBuffer();
		boolean capitalizeNextChar = false;

		//proceeding with the sanitization
		name = name.trim();
		for (int i = 0; i < name.length(); i++) {
			char ch = name.charAt(i);
			if ((i == 0) && !isValidFirstChar(ch)) {
				result.append('_');
			} else if (((i == 0) && isValidFirstChar(ch)) || ((i > 0) && isValidChar(ch))) {
				if (capitalizeNextChar) {
					result.append(Character.toUpperCase(ch));
					capitalizeNextChar = false;
				} else {
					result.append(ch);
				}
			} else if (Character.isWhitespace(ch)) {
				capitalizeNextChar = true;
			}
		}
		return result.toString();
	}

}
