/***********************************************************
 * This software is part of the ProM package * http://www.processmining.org/ * *
 * Copyright (c) 2003-2006 TU/e Eindhoven * and is licensed under the * Common
 * Public License, Version 1.0 * by Eindhoven University of Technology *
 * Department of Information Systems * http://is.tm.tue.nl * *
 **********************************************************/

package org.processmining.plugins.ltlchecker;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Vector;

import javax.swing.JPanel;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XAttributeLiteral;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.Progress;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.plugins.ltlchecker.formulatree.FormulaNode;
import org.processmining.plugins.ltlchecker.formulatree.RootNode;
import org.processmining.plugins.ltlchecker.model.LTLModel;
import org.processmining.plugins.ltlchecker.parser.FormulaParameter;
import org.processmining.plugins.ltlchecker.parser.LTLParser;
import org.processmining.plugins.ltlchecker.parser.ParseException;
import org.processmining.plugins.ltlchecker.util.ParamData;
import org.processmining.plugins.ltlchecker.util.ParamTable;
import org.processmining.plugins.ltlchecker.util.SetsSet;
import org.processmining.plugins.ltlchecker.util.Substitutes;
import org.processmining.plugins.ltlchecker.util.TreeBuilder;

@Plugin(name = "LTL Checker", parameterLabels = { "Log", "LTL Model" }, returnLabels = { "Checking Results",
		"Compliant Traces", "Non-Compliant Traces" }, returnTypes = { CheckResultObject.class, XLog.class, XLog.class }, userAccessible = true)
public class LTLChecker {

	/**
	 * A plugin for the ProM framework to check for properties specified in LTL
	 * on workflow logs of processes. It is an analysis plugin, so it can be
	 * used after mining has been done on the log. Needed for this plugin are a
	 * logReader object and a imported LTL Template file by the import plugin.
	 * 
	 * @version 0.2
	 * @author Fabrizio M. Maggi
	 */

	// FIELDS
	private LTLVerificationResult resultFr;

	/** A table for valuate the parameters. */
	private Hashtable paramTable;

	/** Sets with the sets. */
	private SetsSet sets;

	private UIPluginContext context;
	private JPanel mainPanel;
	private Progress prog;
	private Vector[][] failureType;
	private Vector[][] failurePlace;
	private boolean[][] matrix;
	private RuleModel[] rules;
	private InstanceModel[] instances;

	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "F.M. Maggi", email = "F.M.Maggi@tue.nl", uiLabel = "LTL Checker", pack = "LTLChecker")
	@PluginVariant(variantLabel = "LTL Checker", requiredParameterLabels = { 0, 1 })
	public Object[] analyse(UIPluginContext context, XLog log, LTLModel model) {

		FileInputStream fis = null;
		LTLParser newparser = null;
		File fileLTL = null;

		try {
			fileLTL = File.createTempFile("tempLTL", ".ltl");
			PrintStream out = new PrintStream(fileLTL);
			out.println(model.getFile());
			fis = new FileInputStream(fileLTL.getAbsolutePath());
			newparser = new LTLParser(fis);
			newparser.setFilename(fileLTL.getAbsolutePath());
			newparser.init();
			newparser.parse();
			if (fis != null) {
				fis.close();
			}
		} catch (IOException e) {
			e.printStackTrace();
		} catch (ParseException e) {
			e.printStackTrace();
		}
		CheckVisualizerGUI res = new CheckVisualizerGUI(log, newparser, this);
		JPanel a = res.createGui();
		InteractionResult r = context.showWizard("Configure the LTL Checker", true, true, a);
		if (r == InteractionResult.FINISHED) {
			Vector rules = (Vector) res.getSettings()[0];
			paramTable = (Hashtable) res.getSettings()[1];
			boolean firstSuccess = ((Boolean) res.getSettings()[2]).booleanValue();
			boolean firstFailure = ((Boolean) res.getSettings()[3]).booleanValue();
			boolean skipReady = ((Boolean) res.getSettings()[4]).booleanValue();
			newparser = (LTLParser) res.getSettings()[5];
			prog = context.getProgress();
			prog.setMinimum(0);
			prog.setMaximum(112);
			prog.setIndeterminate(false);
			prog.setValue(3);
			CheckResultObject outputFrame = declareMinerFrame(false, paramTable, log, newparser, rules,
					firstSuccess, firstFailure, skipReady);
			//JComponent outputComponent = new SlickerOpenCheckResults().showLogVis(
				//	outputFrame.getTpRules(), outputFrame.getTpInstances());
			XLog comLog = (XLog) log.clone();
			XLog ncomLog = (XLog) log.clone();
			comLog.clear();
			ncomLog.clear();
			XConceptExtension.instance().assignName(comLog,
					"Compliant traces" + XConceptExtension.instance().extractName(comLog));
			XConceptExtension.instance().assignName(ncomLog,
					"Non Compliant traces" + XConceptExtension.instance().extractName(ncomLog));
			InstanceModel[] insts = outputFrame.getInstances();
			for (int h = 0; h < insts.length; h++) {
				XTrace newT = insts[h].getInstance();
				if (insts[h].isCorrect()) {
					comLog.add(newT);
				} else {
					ncomLog.add(newT);
				}
			}
			//context.addConnection(new CheckVisualizerConnection(outputComponent));
			return new Object[] { outputFrame, comLog, ncomLog };
		}
		context.getFutureResult(0).cancel(true);
		return null;
	}

	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "F.M. Maggi", email = "F.M.Maggi@tue.nl", uiLabel = "LTL Checker Default", pack = "LTLChecker")
	@PluginVariant(variantLabel = "LTL Checker Default", requiredParameterLabels = { 0 })
	public Object[] analyse(UIPluginContext context, XLog log) {

		FileInputStream fis = null;
		LTLParser newparser = null;
		File fileLTL = null;

		try {

			InputStream ir = ClassLoader.getSystemClassLoader().getResourceAsStream("resources/standard.ltl");
			fileLTL = File.createTempFile("tempLTL", ".ltl");
			BufferedReader br = new BufferedReader(new InputStreamReader(ir));
			String line = br.readLine();

			PrintStream out = new PrintStream(fileLTL);
			while (line != null) {
				out.println(line);
				line = br.readLine();
			}
			out.flush();
			out.close();
			fis = new FileInputStream(fileLTL.getAbsolutePath());
			newparser = new LTLParser(fis);
			newparser.setFilename(fileLTL.getAbsolutePath());
			newparser.init();
			newparser.parse();
			if (fis != null) {
				fis.close();
			}
		} catch (IOException e) {
			e.printStackTrace();
		} catch (ParseException e) {
			e.printStackTrace();
		}
		CheckVisualizerGUI res = new CheckVisualizerGUI(log, newparser, this);
		JPanel a = res.createGui();
		InteractionResult r = context.showWizard("Configure the LTL Checker", true, true, a);
		if (r == InteractionResult.FINISHED) {
			Vector rules = (Vector) res.getSettings()[0];
			paramTable = (Hashtable) res.getSettings()[1];
			boolean firstSuccess = ((Boolean) res.getSettings()[2]).booleanValue();
			boolean firstFailure = ((Boolean) res.getSettings()[3]).booleanValue();
			boolean skipReady = ((Boolean) res.getSettings()[4]).booleanValue();
			newparser = (LTLParser) res.getSettings()[5];
			prog = context.getProgress();
			prog.setMinimum(0);
			prog.setMaximum(112);
			prog.setIndeterminate(false);
			prog.setValue(3);
			CheckResultObject outputFrame = declareMinerFrame(false, paramTable, log, newparser, rules,
					firstSuccess, firstFailure, skipReady);
			//JComponent outputComponent = new SlickerOpenCheckResults().showLogVis(
				//	outputFrame.getTpRules(), outputFrame.getTpInstances());
			XLog comLog = (XLog) log.clone();
			XLog ncomLog = (XLog) log.clone();
			comLog.clear();
			ncomLog.clear();
			XConceptExtension.instance().assignName(comLog,
					"Compliant traces" + XConceptExtension.instance().extractName(comLog));
			XConceptExtension.instance().assignName(ncomLog,
					"Non Compliant traces" + XConceptExtension.instance().extractName(ncomLog));
			InstanceModel[] insts = outputFrame.getInstances();
			for (int h = 0; h < insts.length; h++) {
				XTrace newT = insts[h].getInstance();
				if (insts[h].isCorrect()) {
					comLog.add(newT);
				} else {
					ncomLog.add(newT);
				}
			}
			//context.addConnection(new CheckVisualizerConnection(outputComponent));
			return new Object[] { outputFrame, comLog, ncomLog };
		}
		context.getFutureResult(0).cancel(true);
		return null;
	}

	public static void setLabels(UIPluginContext context, XLog log) {
		context.getFutureResult(0).setLabel(
				"Check results (from " + XConceptExtension.instance().extractName(log) + ")");
		context.getFutureResult(1).setLabel(
				"Compliant traces (from " + XConceptExtension.instance().extractName(log) + ")");
		context.getFutureResult(2).setLabel(
				"Non Compliant traces (from " + XConceptExtension.instance().extractName(log) + ")");
	}

	public CheckResultObject declareMinerFrame(boolean console, Hashtable paramTable, XLog log, LTLParser parser,
			Vector enabledRules, boolean firstSuccess, boolean firstFailure, boolean skipReady) {
		Iterator logIt = log.iterator();
		int l = 0;
		int k = 0;
		XLogInfo info = XLogInfoFactory.createLogInfo(log);
		int numTraces = info.getNumberOfTraces();
		int numRules = enabledRules.size();
		instances = new InstanceModel[numTraces];
		XTrace temp = null;
		int i = 0;
		while (logIt.hasNext()) {
			temp = (XTrace) logIt.next();
			instances[i] = new InstanceModel();
			instances[i].setInstance(temp);
			instances[i].setCorrect(true);
			instances[i].setHealthDegree(0);
			i++;
		}
		int badRules = 0;
		int badInstances = 0;
		matrix = new boolean[numRules][numTraces];
		failureType = new Vector[numRules][numTraces];
		failurePlace = new Vector[numRules][numTraces];
		rules = new RuleModel[numRules];
		Substitutes sss = null;
		boolean run = true;
		k = 0;
		while ((k < numRules) && run) {
			//	System.out.println(k);
			if (!console) {
				prog.setValue(50 + (62 * k / enabledRules.size()));
			}
			String formulaName = (String) enabledRules.get(k);
			if (!((formulaName.equals("")) || (formulaName == null))) {
				int teller = 0;
				if (sets == null) {
					sets = new SetsSet(parser, log);
					sets.fill(log);
					//	System.out.println(sets);
					teller = numTraces;
				}
				TreeBuilder tb = new TreeBuilder(parser, formulaName, sets /*
																			 * ,
																			 * log
																			 * .
																			 * getLogSummary
																			 * (
																			 * )
																			 * .
																			 * getOntologies
																			 * (
																			 * )
																			 * null
																			 */);

				// Create the set of substitutes if any. THat is, if
				// it is an template, it can have already substitution
				// now, so add them. Otherwise, an empty substitutes
				// set is enough.

				RootNode root1 = new RootNode();
				sss = (Substitutes) paramTable.get(formulaName);
				sss.setBinder(root1);

				boolean fullfill = false;
				FormulaNode formula1 = (FormulaNode) tb.build(parser.getFormula(formulaName), sss, root1);
				root1.setFormula(formula1);

				rules[k] = new RuleModel();
				rules[k].setRuleName((String) enabledRules.get(k));
				rules[k].setLtlRule(formula1.toString());
				rules[k].setSatisfied(true);
				l = 0;
				while ((l < numTraces) && run) {
					RootNode root = new RootNode();
					sss = (Substitutes) paramTable.get(formulaName);
					sss.setBinder(root);
					FormulaNode formula = (FormulaNode) tb.build(parser.getFormula(formulaName), sss, root);
					root.setFormula(formula);
					failurePlace[k][l] = new Vector();
					failureType[k][l] = new Vector();

					XAttributeLiteral result = null;
					if (instances[l].getInstance().getAttributes() != null) {
						result = (XAttributeLiteral) instances[l].getInstance().getAttributes().get(formula.toString());
					}
					//CHECK
					if ((result != null) && skipReady) {
						fullfill = Boolean.valueOf(result.getValue()).booleanValue();
						matrix[k][l] = fullfill;
					}
					teller++;
					if ((result == null) || !skipReady) {
						// Because the pi must be walked through in reverse order, and the
						// framework does not support this, the ates of this pi are first
						// readed in into a list.

						LinkedList atesList = new LinkedList();
						Iterator ateIt = instances[l].getInstance().iterator();
						while (ateIt.hasNext()) {
							XEvent at = (XEvent) ateIt.next();
							atesList.add(at);
						}
						fullfill = false;
						for (int j = atesList.size(); j >= 0; j--) {
							// start with n + 1 ...
							fullfill = root.value(instances[l].getInstance(), atesList, j);
							RootNode currentRoot = new RootNode();
							FormulaNode currentFormula = (FormulaNode) tb.build(parser.getFormula(formulaName), sss,
									currentRoot);
							currentRoot.setFormula(currentFormula);
						}
						if (fullfill) {
							failurePlace[k][l] = new Vector();
							failureType[k][l] = new Vector();
						}
						matrix[k][l] = fullfill;
					}
					// The computed fullfill of the last ate of the pi is the value of
					// the whole pi, because by dynamic programming the other ates are
					// already computed and used.
					if (!fullfill && firstFailure) {
						// stop at first failure
						run = false;
					} else if (fullfill && firstSuccess) {
						// stop at first success
						run = false;
					}
					if (!fullfill) {
						if (instances[l].isCorrect()) {
							badInstances++;
						}
						if (rules[k].isSatisfied()) {
							badRules++;
						}
						instances[l].setCorrect(false);
						rules[k].setSatisfied(false);
					} else {
						double coverage = instances[l].getHealthDegree();
						coverage++;
						instances[l].setHealthDegree(coverage);
						coverage = rules[k].getCoverage();
						coverage++;
						rules[k].setCoverage(coverage);
					}
					l++;
				}
			}
			double coverage = rules[k].getCoverage();
			coverage = coverage / numTraces;
			if (l < numTraces) {
				coverage = -100;
			}
			rules[k].setCoverage(coverage);
			k++;
		}
		for (int m = 0; m < numTraces; m++) {
			double coverage = instances[m].getHealthDegree();
			coverage = coverage / numRules;
			if (k < numRules) {
				coverage = -100;
			}
			instances[m].setHealthDegree(coverage);
		}
		InstanceModel[] instMod = new InstanceModel[l];
		for (int j = 0; j < l; j++) {
			instMod[j] = instances[j];
		}
		RuleModel[] ruleMod = new RuleModel[k];
		for (int j = 0; j < k; j++) {
			ruleMod[j] = rules[j];
		}
		CheckResultObject crObj = new CheckResultObject();
		crObj.setInstances(instMod);
		crObj.setMatrix(matrix);
		crObj.setRules(ruleMod);
		crObj.setBadInstances(badInstances);
		crObj.setBadRules(badRules);
		crObj.setFailurePlace(failurePlace);
		crObj.setFailureType(failureType);
		//resultFr = new LTLVerificationResult(crObj);
		return crObj;
	}

	protected ParamData createDataModel(List<FormulaParameter> items) {
		return new ParamData(items);
	}

	protected ParamTable createParamTable() {
		return new ParamTable();
	}
}
